# -*- coding: utf8 -*-
from __future__ import unicode_literals

import time
import logging
from driverng import Worklist
from driverng.astm import (
    AstmMessageV2,
    Segment,
    HeaderSegment,
    HprimList,
    HprimString,
    HprimDate,
    HprimForeign,
)

logger = logging.getLogger("driverng.alinity")


class AlinitySegmentMSH(HeaderSegment):
    sending_application = HprimString(3)
    sending_facility = HprimString(4)
    receiving_application = HprimString(5)
    receiving_facility = HprimString(6)
    message_date = HprimDate(7)
    message_type = HprimString((9, 1), default="OML")
    message_code = HprimString((9, 2), default="O33")
    message_label = HprimString((9, 3), default="OML_O33")
    control_id = HprimString(10)
    processing_id = HprimString(11, default="P")
    version_id = HprimString(12, default="2.5.1")
    accept_acknowledgment_type = HprimString(15, default="NE")
    application_acknowledgment = HprimString(16, default="AL")
    character_set = HprimString(18, default="UNICODE UTF-8")
    message_profile = HprimString((21, 1), default="LAB-28")
    message_profile_code = HprimString((21, 2), default="IHE")

    patients = HprimForeign("PID", "header")
    specimens = HprimForeign("SPM", "header")
    orders = HprimForeign("OBR", "header")
    queries = HprimForeign("QPD", "header")
    errors = HprimForeign("ERR", "header")
    equals = HprimForeign("EQU", "header")


class AlinitySegmentINV(Segment):
    substance_identifier = HprimString((3, 1))
    substance_status = HprimString((3, 2))
    substance_type = HprimString((3, 3))
    inventory_container_type = HprimString((4, 2))
    inventory_container_identifier_lot = HprimString((4, 3))
    serial_number = HprimString((5, 1))
    serial_number_coding = HprimString((5, 3))
    expiration_datetime = HprimDate(13)
    first_use_datetime = HprimDate(14)
    # Lot number reagent, control or calibrator material
    manufacturer_lot_number = HprimString(17)


class AlinitySegmentNTE(Segment):
    comment_source = HprimString(3)
    comment_text = HprimString(4)
    comment_type1 = HprimString((5, 1))
    comment_type2 = HprimString((5, 2))
    comment_type3 = HprimString((5, 3))


class AlinitySegmentOBR(Segment):
    placer_order_number = HprimString(3)
    test_identifier = HprimString((5, 1))
    test_text = HprimString((5, 2))
    test_name_coding = HprimString((5, 3), default="99ABT")
    action_code = HprimString(12)
    ordering_provider = HprimString(17)

    results = HprimForeign("OBX", "order")
    infos_tests = HprimForeign("TCD", "order")
    common_order = HprimForeign("ORC", "order")
    sample_priority = HprimForeign("TQ1", "order")

    def to_string(self, separators=None):
        self.rang_segment = ""
        return super(AlinitySegmentOBR, self).to_string(separators=separators)


class AlinitySegmentOBX(Segment):
    value_type = HprimString(3)
    assay_number = HprimString((4, 1))
    result_type_code = HprimString((4, 2))
    observation_sub_id = HprimString(5)
    value = HprimString(6)
    units = HprimString((7, 1))
    units_label = HprimString((7, 2))
    units_coding = HprimString((7, 3))
    reference_range = HprimString(8)
    interpretation_codes = HprimList(9)
    result_status = HprimString(12)
    responsible_observer = HprimString(17)
    equipement_identifier = HprimList(19)
    date_analysis = HprimDate(20)
    observation_type = HprimString(30)

    infos_tests = HprimForeign("TCD", "order")
    infos = HprimForeign("NTE", "result")
    inventory = HprimForeign("INV", "result")


class AlinitySegmentERR(Segment):
    error_location = HprimString(2)
    hl7_error_code = HprimString(3)
    severity = HprimString(4)
    application_error_code = HprimString(5)


class AlinitySegmentEQU(Segment):
    instrument_model = HprimString((2, 1))
    serial_number = HprimString((2, 2))
    serial_name = HprimString((2, 3))
    event_date = HprimDate(3)

    tests_dispo = HprimForeign("TCC", "header")


class AlinitySegmentTCC(Segment):
    test_code = HprimString((2, 1))
    test_label = HprimString((2, 2))
    test_name = HprimString((2, 3))
    code = HprimString(3)
    dispo_code = HprimString((15, 1))
    dispo_label = HprimString((15, 2))
    dispo_name = HprimString((15, 3))


class AlinitySegmentORC(Segment):
    order_control = HprimString(2)
    placer_order_number = HprimString(3)
    order_status = HprimString(5)
    parent = HprimString(8)
    date_transaction = HprimDate(10)

    sample_priority = HprimForeign("TQ1", "patient")
    results = HprimForeign("OBX", "order")

    def to_string(self, separators=None):
        self.rang_segment = ""
        return super(AlinitySegmentORC, self).to_string(separators=separators)


class AlinitySegmentPID(Segment):
    patient_id = HprimString(4)
    lastname = HprimString((6, 1))
    firstname = HprimString((6, 2))
    name_type_code = HprimString((6, 7), default="L")
    birthdate = HprimDate(8, fmt_date="%Y%m%d")
    sex = HprimString(9)

    visits = HprimForeign("PV1", "patient")
    orders = HprimForeign("OBR", "patient")
    infos = HprimForeign("NTE", "patient")
    specimens = HprimForeign("SPM", "patient")

    def to_string(self, separators=None):
        self.rang_segment = ""
        return super(AlinitySegmentPID, self).to_string(separators=separators)


class AlinitySegmentPV1(Segment):
    patient_class = HprimString(3, default="N")
    location = HprimString((4, 2))

    orders = HprimForeign("OBR", "patient")
    specimens = HprimForeign("SPM", "patient")

    def to_string(self, separators=None):
        self.rang_segment = ""
        return super(AlinitySegmentPV1, self).to_string(separators=separators)


class AlinitySegmentSAC(Segment):
    identifier = HprimString(4)
    status = HprimString(9)
    carrier = HprimString(11)
    position = HprimString(12)
    location = HprimString(16)
    dilution = HprimString(30)

    sample_priority = HprimForeign("TQ1", "patient")
    orders = HprimForeign("OBR", "patient")
    common_order = HprimForeign("ORC", "patient")
    inventory = HprimForeign("INV", "result")

    def to_string(self, separators=None):
        self.rang_segment = ""
        return super(AlinitySegmentSAC, self).to_string(separators=separators)


class AlinitySegmentSPM(Segment):
    specimen_type = HprimString(5, default="\"\"")
    role_id = HprimString((12, 1))
    role_label = HprimString((12, 2))
    role_coding_system = HprimString((12, 3))
    collection_date = HprimDate(18)

    orders = HprimForeign("OBR", "patient")
    infos_samples = HprimForeign("SAC", "patient")
    sample_priority = HprimForeign("TQ1", "patient")

    def to_string(self, separators=None):
        self.rang_segment = "1"
        return super(AlinitySegmentSPM, self).to_string(separators=separators)


class AlinitySegmentTCD(Segment):
    universal_service_identifier = HprimString((2, 1))
    test_label = HprimString((2, 2))
    test_coding_name = HprimString((2, 3))
    auto_dilution_type = HprimString((12, 1))
    dilution_coding = HprimString((12, 3))

    orders = HprimForeign("OBR", "patient")
    inventory = HprimForeign("INV", "result")


class AlinitySegmentTQ1(Segment):
    priority_code = HprimString((10, 1))
    priority_label = HprimString((10, 2))
    priority_coding = HprimString((10, 3))

    orders = HprimForeign("OBR", "patient")
    results = HprimForeign("OBX", "order")


class AlinitySegmentQPD(Segment):
    query_name = HprimString((4, 1))

    queries_infos = HprimForeign("RCP", "queries")


class AlinitySegmentRCP(Segment):
    query_priority = HprimString(1)
    response_modality = HprimString(3)


class Driver(AstmMessageV2):
    SPEC = {
        "code": "alinity",
        "label": "ALINITY ABBOTT"
    }

    SEGMENT_CLASS_MAP = {
        "MSH": AlinitySegmentMSH,
        "EQU": AlinitySegmentEQU,
        "TCC": AlinitySegmentTCC,
        "INV": AlinitySegmentINV,
        "NTE": AlinitySegmentNTE,
        "OBR": AlinitySegmentOBR,
        "OBX": AlinitySegmentOBX,
        "ERR": AlinitySegmentERR,
        "ORC": AlinitySegmentORC,
        "PID": AlinitySegmentPID,
        "PV1": AlinitySegmentPV1,
        "SAC": AlinitySegmentSAC,
        "SPM": AlinitySegmentSPM,
        "TCD": AlinitySegmentTCD,
        "TQ1": AlinitySegmentTQ1,
        "QPD": AlinitySegmentQPD,
        "RCP": AlinitySegmentRCP,
    }
    DEFAULT_SEPARATORS = "|^~\\&"
    FIRST_SEP_INDEX = 3
    SEGMENT_ENDING = "\r"

    OPTIONS = [
        AstmMessageV2.Option("alinity_sender_code", label="Code emeteur", default="LISApp"),
        AstmMessageV2.Option("alinity_sender_label", label="Libelle emetteur", default="MainFacility"),
        AstmMessageV2.Option("alinity_receiver_code", label="Code recepteur", default="AlinityApp"),
        AstmMessageV2.Option("alinity_receiver_label", label="Libelle recepteur", default="LabFacility"),
        AstmMessageV2.BooleanOption(
            "alinity_dilution_en_alarme",
            label="Traiter le facteur du dilution comme alarme du résultat"
        )
    ]

    def __init__(
            self,
            alinity_sender_code="LISApp",
            alinity_sender_label="MainFacility",
            alinity_receiver_code="AlinityApp",
            alinity_receiver_label="LabFacility",
            alinity_dilution_en_alarme=False,
            **kwargs
    ):
        self.sender_code = alinity_sender_code
        self.sender_label = alinity_sender_label
        self.receiver_code = alinity_receiver_code
        self.receiver_label = alinity_receiver_label
        self.dilution_en_alarme = alinity_dilution_en_alarme
        super(Driver, self).__init__(**kwargs)

    def create(self, **kwargs):
        return AlinitySegmentMSH(type_segment="MSH", **kwargs)

    def to_string(self, message, separators=None, **kwargs):
        return super(Driver, self).to_string(message, separators, **kwargs) + "\r"

    def encode_worklist(self, worklist, context):
        msg = self.create(
            sending_application=self.sender_code,
            sending_facility=self.sender_label,
            receiving_application=self.receiver_code,
            receiving_facility=self.receiver_label,
            message_date=context["now"],
            control_id=context.get(
                "control_id", int(time.time() * 1000))
        )
        for patient in worklist.get_astm_worklist(
            unique_together=["action", "priority", "rerun"],
            priority={"A": "S", "S": "S", worklist.ELSE: "R"},
            action={"del": "CA", worklist.ELSE: "NW"},
        ):
            seg_pid = msg.patients.add(
                patient_id=patient.patient_id,
                lastname=patient.last_name,
                firstname=patient.first_name,
                birthdate=patient.birthdate,
                sex=patient.sex
            )
            if patient.location:
                seg_pv1 = seg_pid.visits.add(location=patient.location)
            else:
                seg_pv1 = seg_pid
            for sample, action, priority, rerun in patient.samples:
                seq = 1
                for order in sample["orders"]:
                    seg_spm = seg_pv1.specimens.add(
                        role_id="P",
                        role_label="Patient",
                        role_coding_system="HL70369",
                        collection_date=sample.collection_date
                    )
                    seg_sac = seg_spm.infos_samples.add(
                        identifier=sample.sample_id
                    )
                    if not rerun and order.dilution:
                        seg_sac.dilution = order.dilution
                    seg_sac.common_order.add(
                        order_control=action,
                        date_transaction=sample.request_date
                    )
                    if priority == "R":
                        priority_label = "Routine"
                    elif priority == "S":
                        priority_label = "STAT"
                    seg_tq1 = seg_sac.sample_priority.add(
                        priority_code=priority,
                        priority_label=priority_label,
                        priority_coding="HL70485"
                    )
                    seg_obr = seg_tq1.orders.add(
                        placer_order_number=context.get("placer_order_number", int(time.time() * 1000 + seq)),
                        test_identifier=order.code,
                        test_text=order.code,
                    )
                    for physician in sample.physicians:
                        seg_obr.ordering_provider = physician.code
                    if rerun and order.dilution:
                        seg_obr.infos_tests.add(
                            universal_service_identifier=order.code,
                            test_label=order.code,
                            test_coding_name="99ABT",
                            auto_dilution_type=order.dilution,
                            dilution_coding="99ABT",
                        )
                    seq += 1
        if len(msg.patients):
            return msg
        elif context["query"]:
            for sample_id in [seg_qpd.query_name for seg_qpd in context["query"].queries]:
                seg_spm = msg.specimens.add(
                    role_id="U",
                    role_label="Unknow",
                    role_coding_system="HL70369"
                )
                seg_sac = seg_spm.infos_samples.add(identifier=sample_id)
                seg_sac.common_order.add(
                    order_control="DC",
                    date_transaction=context["now"]
                )
            return msg

    def decode_message(self, message):
        if message.queries:
            worklist = Worklist(context="query")
            for seg_qpd in message.queries:
                sample = worklist.get_or_create(
                    type="patient",
                    sample_id=seg_qpd.query_name
                )
        else:
            worklist = Worklist(context="result")
            patients = []
            if message.patients:
                patients = message.patients
            elif message.specimens:
                for seg_spm in message.specimens:
                    if seg_spm.role_id == "Q":
                        sample = worklist.get_or_create(
                            type="control"
                        )
                        for seg_sac in seg_spm.infos_samples:
                            sample.sample_id = seg_sac.identifier
                            sample.control_lot = seg_sac.identifier
                            for seg_obr in seg_sac.orders:
                                if seg_obr.sample_priority:
                                    results = seg_obr.sample_priority[0].results
                                elif seg_obr.common_order:
                                    results = []
                                    if seg_obr.common_order[0].sample_priority:
                                        segs_orc = seg_obr.common_order[0].sample_priority
                                        if segs_orc[0].results:
                                            results = segs_orc[0].results
                                else:
                                    results = seg_obr.results
                                for seg_obx in results:
                                    sample.results.add(
                                        code=seg_obx.assay_number,
                                        value=seg_obx.value,
                                        date=seg_obx.date_analysis,
                                        instrument=seg_obx.equipement_identifier[1],
                                        operator=seg_obx.responsible_observer,
                                        unit=seg_obx.units
                                    )
            for patient in patients:
                if patient.visits:
                    seg_pv1 = patient.visits[0]
                else:
                    seg_pv1 = patient
                for seg_spm in seg_pv1.specimens:
                    for seg_sac in seg_spm.infos_samples:
                        if seg_sac.carrier and seg_sac.position:
                            rack = "{}-{}".format(
                                seg_sac.carrier,
                                seg_sac.position
                            )
                        else:
                            rack = ""
                        sample = worklist.get_or_create(
                            sample_id=seg_sac.identifier,
                            rack=rack,
                            type="patient"
                        )
                        for seg_obr in seg_sac.orders:
                            if seg_obr.common_order:
                                common_order = seg_obr.common_order[0]
                                if seg_obr.common_order[0].sample_priority:
                                    results = common_order.sample_priority[0].results
                                else:
                                    results = seg_obr.results
                                for seg_obx in results:
                                    flags = []
                                    if seg_obx.interpretation_codes:
                                        for flag in seg_obx.interpretation_codes:
                                            if flag != "\"\"":
                                                flags.append(flag)
                                    dilution = ""
                                    for seg_tcd in seg_obx.infos_tests:
                                        if seg_tcd.auto_dilution_type != "STANDARD":
                                            if self.dilution_en_alarme:
                                                flags.append("dilution: {}".format(seg_tcd.auto_dilution_type))
                                            dilution = seg_tcd.auto_dilution_type
                                        else:
                                            dilution = ""
                                    resu = sample.results.add(
                                        code=seg_obx.assay_number,
                                        value=seg_obx.value,
                                        date=seg_obx.date_analysis,
                                        flags=flags,
                                        dilution=dilution,
                                        instrument=seg_obx.equipement_identifier[1],
                                        operator=seg_obx.responsible_observer,
                                        unit=seg_obx.units
                                    )
                                    for seg_tcd in seg_obx.infos_tests:
                                        for seg_inv in seg_tcd.inventory:
                                            if "Reagent" in seg_inv.inventory_container_type:
                                                resu.lots.add(
                                                    type="reagent",
                                                    lot=seg_inv.manufacturer_lot_number,
                                                    expiration=seg_inv.expiration_datetime,
                                                )
                                    for seg_inv in seg_obx.inventory:
                                        if "Reagent" in seg_inv.inventory_container_type:
                                            if seg_inv.serial_number:
                                                reagent_lot = "{}-{}".format(
                                                    seg_inv.manufacturer_lot_number,
                                                    seg_inv.serial_number
                                                )
                                            else:
                                                reagent_lot = seg_inv.manufacturer_lot_number,
                                            resu.lots.add(
                                                type="reagent",
                                                lot=reagent_lot,
                                                expiration=seg_inv.expiration_datetime,
                                            )
        return worklist
